/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package nz.co.gregs.dbvolution;

import java.beans.BeanInfo;
import java.beans.IntrospectionException;
import java.beans.Introspector;
import java.beans.PropertyDescriptor;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import nz.co.gregs.dbvolution.annotations.DBTableColumn;
import nz.co.gregs.dbvolution.annotations.DBTableForeignKey;
import nz.co.gregs.dbvolution.annotations.DBTableName;
import nz.co.gregs.dbvolution.annotations.DBTablePrimaryKey;
import nz.co.gregs.dbvolution.databases.DBDatabase;

/**
 *
 * @author gregory.graham
 */
abstract public class DBTableRow {

    static <T extends DBTableRow> T getInstance(Class<T> requiredDBTableRowClass) {
        try {
            return requiredDBTableRowClass.getConstructor().newInstance();
        } catch (NoSuchMethodException ex) {
            throw new RuntimeException("Unable To Create " + requiredDBTableRowClass.getClass().getSimpleName()+": Please ensure that the constructor of  " + requiredDBTableRowClass.getClass().getSimpleName()+" has no arguments, throws no exceptions, and is public",ex);
        } catch (SecurityException ex) {
            throw new RuntimeException("Unable To Create " + requiredDBTableRowClass.getClass().getSimpleName()+": Please ensure that the constructor of  " + requiredDBTableRowClass.getClass().getSimpleName()+" has no arguments, throws no exceptions, and is public", ex);
        } catch (InstantiationException ex) {
            throw new RuntimeException("Unable To Create " + requiredDBTableRowClass.getClass().getSimpleName()+": Please ensure that the constructor of  " + requiredDBTableRowClass.getClass().getSimpleName()+" has no arguments, throws no exceptions, and is public", ex);
        } catch (IllegalAccessException ex) {
            throw new RuntimeException("Unable To Create " + requiredDBTableRowClass.getClass().getSimpleName()+": Please ensure that the constructor of  " + requiredDBTableRowClass.getClass().getSimpleName()+" has no arguments, throws no exceptions, and is public", ex);
        } catch (IllegalArgumentException ex) {
            throw new RuntimeException("Unable To Create " + requiredDBTableRowClass.getClass().getSimpleName()+": Please ensure that the constructor of  " + requiredDBTableRowClass.getClass().getSimpleName()+" has no arguments, throws no exceptions, and is public", ex);
        } catch (InvocationTargetException ex) {
            throw new RuntimeException("Unable To Create " + requiredDBTableRowClass.getClass().getSimpleName()+": Please ensure that the constructor of  " + requiredDBTableRowClass.getClass().getSimpleName()+" has no arguments, throws no exceptions, and is public", ex);
        }
    }

    private DBDatabase database;
    private List<Field> ignoredRelationships = new ArrayList<Field>();
    private final List<Field> fkFields = new ArrayList<Field>();

    public DBTableRow() {
    }

    public Long getPrimaryKeyLongValue() {
        Long pkColumnValue = -1L;
        boolean pkFound = false;
        QueryableDatatype queryableValueOfField;
        @SuppressWarnings("unchecked")
        Class<? extends DBTableRow> thisClass = this.getClass();
        Field[] fields = thisClass.getDeclaredFields();
        for (Field field : fields) {
            if (field.isAnnotationPresent(DBTablePrimaryKey.class)) {
                pkFound = true;
                queryableValueOfField = this.getQueryableValueOfField(field);
                pkColumnValue = queryableValueOfField.longValue();
                break;
            }
        }
        if (pkColumnValue == null) {
            if (!pkFound) {
                throw new RuntimeException("Primary Key Field Not Defined: Please define the primary key field of " + this.getClass().getSimpleName() + " using the @DBTablePrimaryKey annotation.");
            } else {
                throw new RuntimeException("Primary Key Field Not Parsable as an Integer type or is Null. Please check the PK values of " + this.getClass().getSimpleName());
            }
        } else {
            return pkColumnValue;
        }

    }

    public String getPrimaryKeyStringValue() {
        String pkColumnValue = "";
        boolean pkFound = false;
        QueryableDatatype queryableValueOfField;
        @SuppressWarnings("unchecked")
        Class<? extends DBTableRow> thisClass = this.getClass();
        Field[] fields = thisClass.getDeclaredFields();
        for (Field field : fields) {
            if (field.isAnnotationPresent(DBTablePrimaryKey.class)) {
                pkFound = true;
                queryableValueOfField = this.getQueryableValueOfField(field);
                pkColumnValue = queryableValueOfField.toString();
                break;
            }
        }
        if (!pkFound) {
            throw new RuntimeException("Primary Key Field Not Defined: Please define the primary key field of " + this.getClass().getSimpleName() + " using the @DBTablePrimaryKey annotation.");
        } else {
            return pkColumnValue;
        }

    }

    public String getPrimaryKeySQLStringValue() {
        String pkColumnValue = "";
        QueryableDatatype queryableValueOfField;
        @SuppressWarnings("unchecked")
        Class<? extends DBTableRow> thisClass = this.getClass();
        Field[] fields = thisClass.getDeclaredFields();
        for (Field field : fields) {
            if (field.isAnnotationPresent(DBTablePrimaryKey.class)) {
                queryableValueOfField = this.getQueryableValueOfField(field);
                pkColumnValue = queryableValueOfField.toSQLString();
                break;
            }
        }
        if (pkColumnValue.isEmpty()) {
            throw new RuntimeException("Primary Key Field Not Defined: Please define the primary key field of " + this.getClass().getSimpleName() + " using the @DBTablePrimaryKey annotation.");
        } else {
            return pkColumnValue;
        }

    }

    public String getPrimaryKeyName() {
//        String pkColumnValue = "";
//        QueryableDatatype queryableValueOfField;
        @SuppressWarnings("unchecked")
        Class<? extends DBTableRow> thisClass = this.getClass();
        Field[] fields = thisClass.getDeclaredFields();
        for (Field field : fields) {
            if (field.isAnnotationPresent(DBTablePrimaryKey.class)) {
                return field.getAnnotation(DBTableColumn.class).value();
            }
        }
        throw new RuntimeException("Primary Key Field Not Defined: Please define the primary key field of " + this.getClass().getSimpleName() + " using the @DBTablePrimaryKey annotation.");
    }

    /**
     * DO NOT USE THIS.
     *
     * @param <Q>
     * @param field
     * @return
     */
    @SuppressWarnings("unchecked")
    public <Q extends QueryableDatatype> Q getQueryableValueOfField(Field field) {
        BeanInfo info = null;
        try {
            info = Introspector.getBeanInfo(this.getClass());
        } catch (IntrospectionException ex) {
            throw new RuntimeException("Unable Retrieve Bean Information: Bean Information Not Found For Class: " + this.getClass().getSimpleName());
        }
        PropertyDescriptor[] descriptors = info.getPropertyDescriptors();
        for (PropertyDescriptor pd : descriptors) {
            if (pd.getName().equals(field.getName())) {
                Method readMethod = pd.getReadMethod();
                if (readMethod != null) {
                    try {
                        return (Q) readMethod.invoke(this);
                    } catch (IllegalAccessException ex) {
                        throw new RuntimeException("GET Method Found But Unable To Access: Please change GET method to public for " + this.getClass().getSimpleName() + "." + field.getName(), ex);
                    } catch (IllegalArgumentException ex) {
                        throw new RuntimeException("GET Method Found But Somehow The Argument Was Illegal: Please ensure the read method of " + this.getClass().getSimpleName() + "." + field.getName() + "  has NO arguments.", ex.getCause());
                    } catch (InvocationTargetException ex) {
                        throw new RuntimeException("GET Method Found But Unable To Access: Please ensure the read method of " + this.getClass().getSimpleName() + "." + field.getName() + "  has NO arguments.", ex.getCause());
                    }
                }
            }
        }
        try {
            // no GET method found so try direct method
            return (Q) field.get(this);
            //throw new UnsupportedOperationException("No Appropriate Get Method Found In " + this.getClass().getSimpleName() + " for " + field.toGenericString());
        } catch (IllegalAccessException ex) {
            throw new RuntimeException("Unable To Access Variable Nor GET Method: Please change protection to public for GET method or field " + this.getClass().getSimpleName() + "." + field.getName(), ex);
        }
    }

    Map<String, QueryableDatatype> getColumnsAndQueryableDatatypes() {
        HashMap<String, QueryableDatatype> columnsAndQDTs = new HashMap<String, QueryableDatatype>();
        String columnName;
        QueryableDatatype qdt;

        Field[] fields = this.getClass().getDeclaredFields();
        for (Field field : fields) {
            if (field.isAnnotationPresent(DBTableColumn.class)) {
                qdt = this.getQueryableValueOfField(field);
                columnName = getDBColumnName(field);
                columnsAndQDTs.put(columnName, qdt);
            }
        }
        return columnsAndQDTs;
    }

    /**
     *
     * @return 
     * 
     */
    public String getWhereClause() {
        StringBuilder whereClause = new StringBuilder();
        Field[] fields = this.getClass().getDeclaredFields();
        for (Field field : fields) {
            if (field.isAnnotationPresent(DBTableColumn.class)) {
                QueryableDatatype qdt = this.getQueryableValueOfField(field);
                qdt.setDatabase(this.database);
                whereClause.append(qdt.getWhereClause(this.database.formatTableAndColumnName(this.getTableName(), getDBColumnName(field))));
            }
        }
        return whereClause.toString();
    }

    /**
     * Probably not needed by the programmer, this is the convenience function
     * to find the table name specified by
     *
     * @DBTableName
     *
     * @return the name of the table in the database specified to correlate with
     * the specified type
     *
     */
    public String getTableName() {
        @SuppressWarnings("unchecked")
        Class<? extends DBTableRow> thisClass = (Class<? extends DBTableRow>) this.getClass();
        if (thisClass.isAnnotationPresent(DBTableName.class)) {
            DBTableName annotation = thisClass.getAnnotation(DBTableName.class);
            return annotation.value();
        } else {
            return thisClass.getSimpleName();
        }
    }

    @Override
    public String toString() {
        StringBuilder string = new StringBuilder();
        Class<? extends DBTableRow> thisClass = this.getClass();
        Field[] fields = thisClass.getDeclaredFields();

        String separator = "";

        for (Field field : fields) {
            if (field.isAnnotationPresent(DBTableColumn.class)) {
                string.append(separator);
                string.append(" ");
                string.append(field.getName());
                string.append(":");
                string.append(getQueryableValueOfField(field));
                separator = ",";
            }
        }
        return string.toString();
    }

    /**
     * @param database the database to set
     */
    protected void setDatabase(DBDatabase theDatabase) {
        this.database = theDatabase;

        for (Field field : this.getClass().getDeclaredFields()) {
            if (field.isAnnotationPresent(DBTableColumn.class)) {
                getQueryableValueOfField(field).setDatabase(database);
            }
        }
    }

    protected String getValuesClause() {
        StringBuilder string = new StringBuilder();
        Class<? extends DBTableRow> thisClass = this.getClass();
        Field[] fields = thisClass.getDeclaredFields();

        String separator = " VALUES ( ";
        for (Field field : fields) {
            if (field.isAnnotationPresent(DBTableColumn.class)) {
                string.append(separator).append(getQueryableValueOfField(field).toSQLString());
                separator = ",";
            }
        }
        return string.append(")").toString();
    }

    protected String getSetClause() {
        StringBuilder sql = new StringBuilder();
        Class<? extends DBTableRow> thisClass = this.getClass();
        Field[] fields = thisClass.getDeclaredFields();

        String separator = database.getStartingSetSubClauseSeparator();
        for (Field field : fields) {
            if (field.isAnnotationPresent(DBTableColumn.class)) {
                String columnName = getDBColumnName(field);
                sql.append(separator)
                        .append(database.formatColumnName(columnName))
                        .append(database.getEqualsComparator())
                        .append(getQueryableValueOfField(field)
                        .toSQLString());
                separator = database.getSubsequentSetSubClauseSeparator();
            }
        }
        return sql.toString();
    }

    /**
     *
     * @return 
     */
    protected List<String> getColumnNames() {
        ArrayList<String> columnNames = new ArrayList<String>();
        Class<? extends DBTableRow> thisClass = this.getClass();
        Field[] fields = thisClass.getDeclaredFields();

        for (Field field : fields) {
            String dbColumnName = getDBColumnName(field);
            if (dbColumnName != null) {
                columnNames.add(dbColumnName);
            }
        }
        return columnNames;
    }

    public String getDBColumnName(Field field) {
        String columnName = "";

        if (field.isAnnotationPresent(DBTableColumn.class)) {
            DBTableColumn annotation = field.getAnnotation(DBTableColumn.class);
            columnName = annotation.value();
            if (columnName
                    == null || columnName.isEmpty()) {
                columnName = field.getName();
            }
        }
        return columnName;
    }

    protected Map<DBTableForeignKey, DBTableColumn> getForeignKeys() {
        HashMap<DBTableForeignKey, DBTableColumn> foreignKeys;
        foreignKeys = new HashMap<DBTableForeignKey, DBTableColumn>();
        Class<? extends DBTableRow> thisClass = this.getClass();
        List<Field> fields = this.getForeignKeyFields();

        for (Field field : fields) {
            if (!ignoredRelationships.contains(field)) {
                DBTableForeignKey annotation = field.getAnnotation(DBTableForeignKey.class);
                DBTableColumn columnName = field.getAnnotation(DBTableColumn.class);
                foreignKeys.put(annotation, columnName);
            }
        }
        return foreignKeys;
    }

    protected List<Field> getForeignKeyFields() {
        if (fkFields.isEmpty()) {
            Class<? extends DBTableRow> thisClass = this.getClass();
            Field[] fields = thisClass.getDeclaredFields();

            for (Field field : fields) {
                if (field.isAnnotationPresent(DBTableForeignKey.class)) {
                    fkFields.add(field);
                }
            }
        }
        return fkFields;
    }

    public Field getFieldOf(QueryableDatatype qdt) {
        Field fieldReqd = null;

        Field[] fields = this.getClass().getDeclaredFields();

        for (Field field : fields) {
            final Object fieldOfThisInstance;
            try {
                fieldOfThisInstance = field.get(this);
            } catch (IllegalArgumentException ex) {
                        throw new RuntimeException("Field Found But Somehow The Argument Was Illegal: Please ensure the fields of " + this.getClass().getSimpleName() + "." + field.getName() + "  are public.", ex.getCause());
            } catch (IllegalAccessException ex) {
                        throw new RuntimeException("Field Found But Unable To Access: Please ensure the fields of " + this.getClass().getSimpleName() + "." + field.getName() + "  are public.", ex);
            }
            if (fieldOfThisInstance.equals(qdt)) {
                return field;
            }
        }
        return fieldReqd;
    }

    /**
     *
     * Requires the field to be from this instance to work
     *
     * @param qdt
     */
    public void ignoreForeignKey(QueryableDatatype qdt) {
        Field fieldOfFK = getFieldOf(qdt);
        ignoredRelationships.add(fieldOfFK);
    }

    public void useAllForeignKeys() {
        ignoredRelationships.clear();
    }

    public void ignoreAllForeignKeys() {
        ignoredRelationships.addAll(this.getForeignKeyFields());
    }
}